//
// by S. Gehlbach <steve @ kesa . com>
//
// edit with tab stops =4
//
// for SiS630 spd interface
//
#ifndef COMMON
#include "common.h"
#endif

#define SMA_SIZE 0x80
#define ACPI_BASE_ADDR		0x5000
#define SMB_BASE_ADDR ACPI_BASE_ADDR

/* Define register offsets */
#define SMB_STS     0x80
#define SMB_ENB     0x81
#define SMB_CNT     0x82
#define SMBHOST_CNT 0x83
#define SMB_ADDR    0x84
#define SMB_CMD     0x85
#define SMB_PCOUNT  0x86
#define SMB_COUNT   0x87
#define SMB_DATA    0x88

/* Define register settings */
#define SMB_KILL    0x20
#define DIMM_BASE   0xa1        // 1010001 is base for DIMM in SMBus
#define READ_CMD    0x12        // start + R/W in CMD reg

/* Define SPD Data locations */
#define MEM_TYPE  2           // Memory Type - EDO, FPM, SDRAM
#define NUM_ROWS  3           // Number of Row Addresses
#define NUM_COLS  4           // Number of Column Addresses
#define NUM_MOD_ROWS 5        // Number of Module Rows
#define NUM_BANKS 17          // number of banks
#define CAS_LAT   18          // CAS Latencies Supported
#define MOD_ATTR  21          // SDRAM Module Attributes
#define BANK_DENSITY 31       // Module Bank Density
#define SPD_REV   62          // SPD Revision
#define SPEC_FREQ 126         // Specification Frequency - 66, 100, ...


// one bank
const char sdram_type_bank_1 [] = {
// Column Number   8       9       10      11    Row Number
                  0x00, 0x04, 0x08, 0xff,  // 11
				  0xff, 0xff, 0xff, 0xff,  // 12
				  0x01, 0x05, 0x09, 0x0d   // 13
};

// two banks
const char sdram_type_bank_2 [] = {
// Column Number   8       9       10      11    Row Number
                  0xc0, 0xff, 0xff, 0xff, // 11
				  0x02, 0x06, 0x0a, 0x0e, // 12
				  0x03, 0x07, 0x0b, 0x0f  // 13
};
extern inline unsigned char read_spd(unsigned char byte_number) {
	outb(0xff,SMB_BASE_ADDR + SMB_STS); // clear status
	outb(byte_number, SMB_BASE_ADDR + SMB_CMD); // SMBus command
	outb(READ_CMD, SMB_BASE_ADDR + SMBHOST_CNT); // SMBus Host Control, Start, R/W
	while ( !(0x0a & inb(SMB_BASE_ADDR + SMB_STS)) ) ; // loop on status for error or done
	return inb(SMB_BASE_ADDR + SMB_DATA);
}

extern inline set_ram (unsigned char slot) {
	// issue smbus kill command
	outb(SMB_KILL, SMB_BASE_ADDR + SMBHOST_CNT); // SMBus host control
	outb(0xff,SMB_BASE_ADDR + SMB_STS); // clear status

	// address the bus
	// SPD is SMBus address 1010 xxxR where R=1 for read and xxx is dimm slot number
	outb(0xa1 + 2*slot,SMB_BASE_ADDR + SMB_ADDR); // SMBus Address

	// read ram type  (SPD byte 2)
	// skip if not SDRAM (type=4)
	if ( read_spd(MEM_TYPE) != 0x04 ) return;

	// above should fail on non-ex simm but just to be sure...
	// check the status
	if (inb(SMB_BASE_ADDR + SMB_STS) != 0x08) return;

	// need to ck and see if 0xff
	switch (read_spd(NUM_ROWS)) {
		case 11:
		case 12:
		case 13:
			break;
		default:
			// not in range, return
			return;
	}
	switch (read_spd(NUM_COLS)) {
		case 8:
		case 9:
		case 10:
		case 11:
			break;
		default:
			// not in range, return
			return;
	}

	//
	// note we read the spd more than once but we are using it as our scratch area
	// since we have no scratch space just registers.
	// some table values are out of range but we will check them later and not enable ram.
	// 
	switch (read_spd(NUM_BANKS)) {

		case 1: pci_conf1_write_config_byte(
					sdram_type_bank_1[read_spd(NUM_ROWS)*4 + read_spd(NUM_COLS) - 52],
					CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x60+slot)
				);
				break;
		default: 
				pci_conf1_write_config_byte(
					sdram_type_bank_2[read_spd(NUM_ROWS)*4 + read_spd(NUM_COLS) - 52],
					CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x60+slot)
				);
	}

	//
	// check illegal values
	// return on error does not enable dimm so effectively it is disabled
	//
	switch (pci_conf1_read_config_byte(CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x60+slot)) & 0x80) {
		case 0x80:
			return;
		default:
	}

	// set module rows
	switch (read_spd(NUM_MOD_ROWS)) {
		case 2:
				pci_conf1_write_config_byte(
					1<<5 | 
					pci_conf1_read_config_byte(CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x60+slot)),
					CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x60+slot)
				);
	}
	//
	// everything okay so set the enable bit
	//
	pci_conf1_write_config_byte(
				(1<<slot) |
				pci_conf1_read_config_byte(CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x63)),
				CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x63)
			);
}

extern inline void setup_chip () {
	// ACPI Enable
	pci_conf1_write_config_byte(
		0x8a,
		CONFIG_CMD(LPC_BRIDGE_BUS,PCI_DEVFN(LPC_BRIDGE_DEV,LPC_BRIDGE_FUN),0x40)
		);
	// Set ACPI Base I/O Address
	pci_conf1_write_config_byte(
		ACPI_BASE_ADDR>>16,
		CONFIG_CMD(LPC_BRIDGE_BUS,PCI_DEVFN(LPC_BRIDGE_DEV,LPC_BRIDGE_FUN),0x75)
		);
	// MDOE# enable, this bit should be set before sizing
	pci_conf1_write_config_byte(
		0x01,
		CONFIG_CMD(LPC_BRIDGE_BUS,PCI_DEVFN(LPC_BRIDGE_DEV,LPC_BRIDGE_FUN),0x55)
		);
	// initialize the dimm register
	pci_conf1_write_config_byte(SMA_SIZE, CONFIG_CMD(NORTH_BRIDGE_BUS,PCI_DEVFN(NORTH_BRIDGE_DEV,NORTH_BRIDGE_FUN),0x63));
	// check each dimm slot
	set_ram(0);
	set_ram(1);
	set_ram(2);
}
